@@ -8,6 +8,7 @@ import shutil
 import warnings
 from altgraph import Graph
 import textwrap
+import xml.etree.ElementTree as ET
 
 try:
     bytes
@@ -78,7 +79,7 @@ class TestFunctions (unittest.TestCase):
                     if self._namespace_packages is None:
                         raise ValueError("no file")
 
-                    return self._namespace_packages 
+                    return self._namespace_packages
 
                 raise ValueError("invalid lookup key")
 
@@ -163,6 +164,8 @@ class TestFunctions (unittest.TestCase):
                 self.assertEqual(filename, os.path.join(path, 'mymodule2.pyc'))
                 self.assertEqual(description, ('.pyc', 'rb', imp.PY_COMPILED))
 
+                fp.close()
+
             # Compiled plain module, with source
 #            info = modulegraph.find_module('mymodule3', path=[path] + sys.path)
 #
@@ -202,19 +205,22 @@ class TestFunctions (unittest.TestCase):
                 if sys.platform == 'win32':
                     ext = '.pyd'
                 else:
-                    # This is a ly, but is good enough for now  
+                    # This is a ly, but is good enough for now
                     ext = '.so'
 
-                self.assertEqual(fp, None)
                 self.assertEqual(filename, os.path.join(path, 'myext' + ext))
                 self.assertEqual(description, (ext, 'rb', imp.C_EXTENSION))
+                self.assertEqual(fp, None)
 
     def test_moduleInfoForPath(self):
         self.assertEqual(modulegraph.moduleInfoForPath("/somewhere/else/file.txt"), None)
 
         info = modulegraph.moduleInfoForPath("/somewhere/else/file.py")
         self.assertEqual(info[0], "file")
-        self.assertEqual(info[1], "U")
+        if sys.version_info[:2] >= (3,4):
+            self.assertEqual(info[1], "r")
+        else:
+            self.assertEqual(info[1], "U")
         self.assertEqual(info[2], imp.PY_SOURCE)
 
         info = modulegraph.moduleInfoForPath("/somewhere/else/file.pyc")
@@ -241,7 +247,7 @@ class TestFunctions (unittest.TestCase):
                 saved_replace = modulegraph.replacePackage
                 try:
                     called = []
-                    
+
                     def log_add(*args, **kwds):
                         called.append(('add', args, kwds))
                     def log_replace(*args, **kwds):
@@ -360,19 +366,25 @@ class TestNode (unittest.TestCase):
         self.assertEqual(d[n2], 'n2')
 
     def test_infoTuple(self):
-         n = modulegraph.Node('n1')
-         self.assertEqual(n.infoTuple(), ('n1',))
+        n = modulegraph.Node('n1')
+        self.assertEqual(n.infoTuple(), ('n1',))
 
     def assertNoMethods(self, klass):
         d = dict(klass.__dict__)
         del d['__doc__']
         del d['__module__']
+        if '__qualname__' in d:
+            # New in Python 3.3
+            del d['__qualname__']
         self.assertEqual(d, {})
 
     def assertHasExactMethods(self, klass, *methods):
         d = dict(klass.__dict__)
         del d['__doc__']
         del d['__module__']
+        if '__qualname__' in d:
+            # New in Python 3.3
+            del d['__qualname__']
 
         for nm in methods:
             self.assertTrue(nm in d, "%s doesn't have attribute %r"%(klass, nm))
@@ -469,7 +481,7 @@ class TestModuleGraph (unittest.TestCase):
 
         g = Graph.Graph()
         o = modulegraph.ModuleGraph(['a', 'b', 'c'], ['modA'], [
-                ('fromA', 'toB'), ('fromC', 'toD')], 
+                ('fromA', 'toB'), ('fromC', 'toD')],
                 {
                     'modA': ['modB', 'modC'],
                     'modC': ['modE', 'modF'],
@@ -562,7 +574,8 @@ class TestModuleGraph (unittest.TestCase):
         graph = modulegraph.ModuleGraph()
         graph.import_hook('os.path', None)
         graph.import_hook('idlelib', None)
-        
+        graph.import_hook('xml.dom', None)
+
         for node in graph.nodes():
             if isinstance(node, modulegraph.Package):
                 break
@@ -582,10 +595,10 @@ class TestModuleGraph (unittest.TestCase):
         #self.assertTrue(m is parent)
 
         m = graph.findNode('xml')
-        self.assertEqual(graph.determine_parent(m), None)
+        self.assertEqual(graph.determine_parent(m), m)
 
         m = graph.findNode('xml.dom')
-        self.assertEqual(graph.determine_parent(m), graph.findNode('xml'))
+        self.assertEqual(graph.determine_parent(m), graph.findNode('xml.dom'))
 
 
     @expectedFailure
@@ -640,7 +653,7 @@ class TestModuleGraph (unittest.TestCase):
                 ('sub1', 'root.sub.sub1', node),
                 ('sub2', 'root.sub.sub1.sub2', node2),
             ])
-        
+
         self.assertRaises(ImportError, graph.load_tail, root, 'raises')
         self.assertRaises(ImportError, graph.load_tail, root, 'sub.raises')
         self.assertRaises(ImportError, graph.load_tail, root, 'sub.raises.sub')
@@ -651,7 +664,7 @@ class TestModuleGraph (unittest.TestCase):
     def test_ensure_fromlist(self):
         # 1. basic 'from module import name, name'
         # 2. 'from module import *'
-        # 3. from module import os 
+        # 3. from module import os
         #    (where 'os' is not a name in 'module',
         #     should create MissingModule node, and
         #     should *not* refer to the global os)
@@ -702,7 +715,7 @@ class TestModuleGraph (unittest.TestCase):
         code = compile(textwrap.dedent('''\
             import sys
             import os.path
-            
+
             def testfunc():
                 import shutil
             '''), '<test>', 'exec', 0, False)
@@ -715,7 +728,7 @@ class TestModuleGraph (unittest.TestCase):
         # from module import *
         #  both:
         #   -> with/without globals
-        #   -> with/without modules in globals (e.g, 
+        #   -> with/without modules in globals (e.g,
         #       from os import * adds dependency to os.path)
         # from .module import a
         # from ..module import a
@@ -772,19 +785,22 @@ class TestModuleGraph (unittest.TestCase):
                 srcfn = srcfn[:-1]
             self.assertEqual(m[1], srcfn)
             self.assertEqual(m[2], ('.py', 'rU', imp.PY_SOURCE))
+            m[0].close()
 
             m2 = graph.find_module('shutil', None)
             self.assertEqual(m[1:], m2[1:])
+            m2[0].close()
 
 
             record[:] = []
             m = graph.find_module('sax', xml.packagepath, xml)
             self.assertEqual(m,
-                    (None, os.path.join(os.path.dirname(xml.filename), 'sax'), 
+                    (None, os.path.join(os.path.dirname(xml.filename), 'sax'),
                     ('', '', imp.PKG_DIRECTORY)))
             self.assertEqual(record, [
                 ('sax', xml.packagepath),
             ])
+            if m[0] is not None: m[0].close()
 
         finally:
             modulegraph.find_module = saved_finder
@@ -829,7 +845,7 @@ class TestModuleGraph (unittest.TestCase):
                     expected.append([type(n).__name__, n.identifier, n.filename])
                 else:
                     expected.append([type(n).__name__, n.identifier])
-                    
+
             expected.sort()
             actual = [item.split() for item in lines[3:]]
             actual.sort()
@@ -927,7 +943,61 @@ class TestModuleGraph (unittest.TestCase):
         self.assertIsInstance(e, int)
         self.assertEqual(graph.graph.edge_data(e), 'direct')
 
+    def test_create_xref(self):
+        # XXX: This test is far from optimal, it just ensures
+        # that all code is exercised to catch small bugs and
+        # py3k issues without verifying that the code actually
+        # works....
+        graph = modulegraph.ModuleGraph()
+        if __file__.endswith('.py'):
+            graph.run_script(__file__)
+        else:
+            graph.run_script(__file__[:-1])
 
+        graph.import_hook('os')
+        graph.import_hook('xml.etree')
+        graph.import_hook('unittest')
+
+        fp = StringIO()
+        graph.create_xref(out=fp)
+
+        data = fp.getvalue()
+        r = ET.fromstring(data)
+
+    def test_itergraphreport(self):
+        # XXX: This test is far from optimal, it just ensures
+        # that all code is exercised to catch small bugs and
+        # py3k issues without verifying that the code actually
+        # works....
+        graph = modulegraph.ModuleGraph()
+        if __file__.endswith('.py'):
+            graph.run_script(__file__)
+        else:
+            graph.run_script(__file__[:-1])
+        graph.import_hook('os')
+        graph.import_hook('xml.etree')
+        graph.import_hook('unittest')
+        graph.import_hook('distutils.command.build')
+
+        fp = StringIO()
+        list(graph.itergraphreport())
+
+        # XXX: platpackages isn't implemented, and is undocumented hence
+        # it is unclear what this is inteded to be...
+        #list(graph.itergraphreport(flatpackages=...))
+
+
+
+
+class CompatTests (unittest.TestCase):
+    def test_Bchr(self):
+        v = modulegraph._Bchr(ord('A'))
+        if sys.version_info[0] == 2:
+            self.assertTrue(isinstance(v, bytes))
+            self.assertEqual(v, b'A')
+        else:
+            self.assertTrue(isinstance(v, int))
+            self.assertEqual(v, ord('A'))
 
 if __name__ == "__main__":
     unittest.main()
